import logging
import time
from asyncio.exceptions import CancelledError
from datetime import datetime
from typing import Callable, Dict, Union

import dask
from dask.distributed import Client
from IPython.display import clear_output, display
from stratosphere import config
from stratosphere.run import Run
from stratosphere.store import serialization
from stratosphere.utils import log


def log_event(msg: Union[Dict, str]):
    """Log structured event with Dask.

    Args:
        msg (Union[Dict, str]): Event to log.
    """
    dask.distributed.get_worker().log_event("stratosphere", msg)


def dask_logger(run: Run, insert_delay: float = 0) -> Run:
    """Dask logger function that can be appended on experiments,
    to log the attributes of executed runs with Dask, enabling
    real-time experiment tracking. Tracked properties:

    Args:
        run (Run): Run to track.
        insert_delay (float, optional): Introduce a delay (seconds),
        useful to demonstrate the tracking functionality. Defaults to 0.

    Returns:
        Run: unaltered run after tracking it.
    """
    if insert_delay > 0:
        time.sleep(insert_delay)

    # Dask doesn't know how to serialize the uuid.UUID type, Pandas data frames, etc.0
    log_event({"id_run": str(run.id_run), "fields": serialization.serialize(run.fields)})
    return run


@log.default_exception_handler
def monitor_latest_event(client_address: str = config.dask.scheduler_address):
    """Utility function to report the latest tracked Run event, useful to start
    debugging execution issues. The function returns after a keyboard interrupt
    or after the connection with the scheduler is lost, which can happen if:
    1. The connection delay is too low (see config.dask)
    2. The Dask scheduler terminated, either nicely or after a failure

    Args:
        client_address (str, optional): Dask scheduler address. Defaults to config.dask.scheduler_address.
    """
    log.init_logging(logging.INFO)

    def display_progress(events):
        clear_output(wait=True)
        print(f"Time: {datetime.now().strftime('%d/%m/%Y %H:%M:%S')} Count events: {len(events)}\n")  # noqa
        # Display fields of last received record
        last_run_fields = serialization.deserialize(events[-1][1]["fields"])
        display(last_run_fields)

    monitor_events(display_progress, client_address=client_address)


@log.default_exception_handler
def monitor_events(func: Callable, client_address: str = config.dask.scheduler_address):
    """Utility function to execute a callable, passing a Pandas dataframe as argument, representing
    all the tracked events. It can be used to implement custom tracking monitors.

    Args:
        func (Callable): Function that accepts as argument a Pandas dataframe, representing the
        tracked events.
        client_address (str, optional): _description_. Defaults to config.dask.scheduler_address.
    """
    try:
        with Client(address=client_address, timeout=config.dask.client_timeout) as client:
            while True:
                events = client.get_events("stratosphere")
                func(events)
                time.sleep(1)
    except (CancelledError, OSError):
        logging.error("Connection failed or cluster terminated")



from typing import Callable, List, Union

from stratosphere.store.db.database import next_ulid


class Run:
    """A run represents an instance of the experiment, obtained by
    combining the fixed and variable parameters. The Run objects
    (and the tracked fields)  must be serializable with cloudpickle.
    """

    def __init__(
        self,
        id_run: str = None,
        funcs: Union[Callable, List[Callable]] = None,
        kwargs: dict = None,
        params: dict = None,
        fields: dict = None,
    ):
        """Create a new run.

        Args:
            kwargs (dict, optional): Fixed parameters for all runs of an experiment. Defaults to None.
            parameters (dict, optional): Variable parameters to be considered for this run. Defaults to None.
            fields (dict, optional): fields to be tracked. Defaults to None.
            funcs (Union[Callable, List[Callable]], optional): One or more functions to be executed. Their
            only parameter is an instance of the Run class itself.
            experiment_name (str, optional): Name of the experiment the run belongs to. Defaults to None.
        """

        self.id_run = next_ulid() if id_run is None else id_run
        self.funcs = normalize_funcs(funcs)
        self.kwargs = {} if kwargs is None else kwargs
        self.params = {} if params is None else params
        self.fields = {} if fields is None else fields
        self.exception = None

    def execute(self, funcs=None):
        if funcs is not None:
            self.funcs = normalize_funcs(funcs)

        self.exception = None

        for func in self.funcs:
            try:
                func(self)
            except Exception as e:  # noqa
                # We want to catch all exceptions in the distributed execution of runs,
                # It will then be reported by Experiment.execute.
                self.exception = f"{func.__qualname__.split('.')[0]}: {str(e)}"
                break

        return self

    def verify(self, func):
        func(self)


def normalize_funcs(funcs):
    if funcs is None:
        return []
    elif callable(funcs):
        return [funcs]
    else:
        return funcs



from stratosphere.run import Run
from stratosphere.utils.assertions import assert_df, assert_keys


def eval_metrics(run: Run):
    run.verify(assert_df(name="fields", key="predictions", columns=["y_true", "y_pred"]))

    metrics = {"accuracy": (run.fields["predictions"]["y_true"] == run.fields["predictions"]["y_pred"]).mean()}

    run.fields["metrics"] = metrics



import json
from typing import List
import uuid
import zlib
from pickle import PickleError

import cloudpickle
import numpy as np
import pandas as pd
from stratosphere import config


def compress(data: bytes, enable_compression=config.serialization.enable_compression) -> bytes:
    if enable_compression:
        return zlib.compress(data)
    else:
        return data


def decompress(data: bytes) -> bytes:
    if isinstance(data, memoryview):
        data = data.tobytes()

    try:
        data = zlib.decompress(data)
    except zlib.error:
        pass

    return data


# Introduced with Python 3.0, https://peps.python.org/pep-3154/
# Unpickling might fail: on different architectures, different python version, in case of missing packages.
PICKLE_DEFAULT_PROTOCOL = 4


def pickle_dumps(obj: object) -> bytes:
    """It returns the compressed (zlib) pickled object. The Pickle DEFAULT_PROTOCOL is used
    for maximum compatibility.

    Args:
        obj (object): Object to serialize.

    Returns:
        bytes: Compressed serialized object.
    """
    return compress(cloudpickle.dumps(obj, protocol=PICKLE_DEFAULT_PROTOCOL))


def pickle_loads(data: bytes) -> object:
    """It returns the loaded object, afer uncompressing and unpickling it.

    Args:
        data (bytes): Serialized object.

    Returns:
        object: Unserialized object.
    """
    try:
        return cloudpickle.loads(decompress(data))
    except Exception:  # noqa # we do want to catch all errors here
        raise PickleError()


def pickle_size(obj: object, unit: str = "b") -> int:
    """It returns the size of the object once serialised (including compression).

    Args:
        obj (object): Object to analyse.
        unit (str, optional): Unit of measure, b (Bytes), kb (KiloBytes), mb (MegaBytes). Defaults to "b".

    Returns:
        int: Size of the serialized object.
    """
    size_object = len(pickle_dumps(obj))
    if unit == "b":
        return int(size_object * 1e2) / 1e2
    elif unit == "kb":
        return int(size_object * 1e-3 * 1e2) / 1e2
    elif unit == "mb":
        return int(size_object * 1e-6 * 1e2) / 1e2
    else:
        return None


serialized_types = [pd.DataFrame, pd.Series, np.ndarray, uuid.UUID, dict, list]


class JSONEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, pd.DataFrame):
            return {"__type": "pandas.DataFrame", "data": obj.to_dict(orient="list"), "index": obj.index.tolist()}
        elif isinstance(obj, pd.Series):
            return {"__type": "pandas.Series", "data": obj.tolist(), "index": obj.index.tolist()}
        elif isinstance(obj, np.ndarray):
            return {"__type": "numpy.ndarray", "data": obj.tolist()}
        elif isinstance(obj, uuid.UUID):
            return {"__type": "uuid.UUID", "data": str(obj)}
        else:
            return json.JSONEncoder.default(self, obj)


def serialize(obj: object, enable_compression=config.serialization.enable_compression) -> bytes:
    return compress(json.dumps(obj, cls=JSONEncoder).encode("UTF-8"), enable_compression=enable_compression)


def deserialize(obj: bytes) -> object:
    def f(d):
        if isinstance(d, dict) and "__type" in d:
            # This ia a value to deserialize
            if d["__type"] == "pandas.DataFrame":
                df = pd.DataFrame.from_dict(d["data"], orient="columns")
                df.index = pd.Index(d["index"])
                return df
            elif d["__type"] == "pandas.Series":
                return pd.Series(d["data"], index=d["index"])
            elif d["__type"] == "numpy.ndarray":
                return np.asarray(d["data"])
            elif d["__type"] == "uuid.UUID":
                return uuid.UUID(d["data"])
            else:
                # We don't know how to deserialize it, return it as a dictionary.
                return d

        if isinstance(d, list):
            # Walk thru the list, trying to decode values.
            return [f(v) for v in d]

        if isinstance(d, dict):
            # Walk thru the dict, trying to decode values.
            return {kv[0]: f(kv[1]) for kv in d.items()}

        else:
            # Nothing to do, return value.
            return d

    return json.loads(decompress(obj).decode("UTF-8"), object_hook=f)


def serialize_df(df: pd.DataFrame, consider_columns: list = None):
    if len(df) == 0:
        # if no rows ... nothing to do.
        return df, []

    if consider_columns is None:
        consider_columns = df.columns

    # Identify columns to serialize
    serialized_cols = []
    for col_name in consider_columns:
        for serialized_type in serialized_types:
            # We assume that the types in the first row are the same of the other rows.
            if isinstance(df[col_name].iloc[0], serialized_type):
                serialized_cols.append(col_name)
                break

    if len(serialized_cols) > 0:
        # If there are columns to serialize, work on a frame copy.
        df = df.copy()
        for col_name in serialized_cols:
            df[col_name] = df[col_name].map(serialize)

    # Identify columns that haven't been serialized, among the ones to be considered.
    non_serialized_cols = [col_name for col_name in consider_columns if col_name not in serialized_cols]

    columns = {"serialized": serialized_cols, "non_serialized": non_serialized_cols}
    return df, columns


def deserialize_df(df: pd.DataFrame, columns: List[str]):
    if len(columns) == 0:
        return df

    df = df.copy()
    for col_name in columns:
        df[col_name] = df[col_name].map(deserialize)

    return df


def explode_json_column(df: pd.DataFrame, col_name: str, prefix: str = None, suffix: str = None) -> pd.DataFrame:
    """Explode a column in the Pandas dataframe containing a dict to a list of columns.
    Useful to handle the "attributes" column in the "Experiments" table, which contains
    the serialized values as JSON.
    Args:
        df (pd.DataFrame): Pandas dataframe
        col_name (str): Column to process.
        prefix (str, optional): Prefix to consider for all exploded columns.
        suffix (str, optional): In case exploded columns already exist, use this suffix.
        Defaults to _{col_name}.
    Returns:
        pd.DataFrame: Resulting Pandas dataframe.
    """

    if suffix is None:
        suffix = f"_{col_name}"

    # Explode column containing json to multiple columns, in a dataframe.
    df_exploded = pd.json_normalize(df[col_name])

    if prefix is not None:
        df_exploded.columns = [f"{prefix}{col_name}" for col_name in df_exploded.columns]

    return df.drop(columns=[col_name]).merge(
        df_exploded,
        how="left",
        left_index=True,
        right_index=True,
        suffixes=(None, suffix),
    )



from enum import Enum
from typing import Union

# Enum used with if-then situation with insertions
IfExists = Enum("IfExists", ["replace", "fail"])

# Enum used with if-then situations with deletions
IfMissing = Enum("IfMissing", ["ignore", "fail"])


def enforce_enum(x: Union[str, Enum], enum_type: Enum) -> Enum:
    """Convert strings to enums, or do nothing in case of enums.

    Args:
        x (Union[str, Enum]): Object to convert.
        enum_type (Enum): Enum type to consider.

    Returns:
        Enum: Converted input object.
    """
    if type(x) == str:
        return enum_type[x]
    else:
        return x



import logging


class db:
    """Parameters affecting the database management."""

    # The default URL can be used for all database operations.
    default_url = "sqlite:///:memory:"

    # Rads and writes are executed in chunks, that enable
    # r/w progress bars.
    query_read_chunk_size = 1000
    query_write_chunk_size = 1000

    experiments_table = "experiments"
    experiment_table_prefix = "e_"


class serialization:
    enable_compression = False
    store_pickle = True


class tqdm:
    """Parameters controlling how tqdm is used/not used."""

    enable = True


class log:
    """Parameters controlling the built-in logging capabilities."""

    level = logging.INFO

    # If enabled, exceptions are reported in a compact way.
    catch_exceptions = True


class dask:
    """Parameters to handle dask execution backends."""

    scheduler_address = "tcp://127.0.0.1:8786"
    dashboard_address = ":8787"
    scheduler_port = 8786
    client_timeout = "5s"



from sqlalchemy import Column, LargeBinary, String
from sqlalchemy.orm import declarative_base
from sqlalchemy_utils.types.uuid import UUIDType
from stratosphere import config

# Construct a base class for declarative class definitions.
Base = declarative_base()

# UUID type, to be used in model definitions. By passing binary=False, we fall back
# to the string representation of UUIDs if there's no native type  (as in SQLite).
uuid_type = UUIDType(binary=False)


class Experiment(Base):
    """Model representing the index record of an experiment in the database."""

    __tablename__ = config.db.experiments_table
    id_experiment = Column(uuid_type, primary_key=True, default=None)
    name = Column(String, nullable=False, unique=True)
    fields = Column(LargeBinary, nullable=False, default=None)
    properties = Column(LargeBinary, nullable=False, default=None)
    pickle = Column(LargeBinary, nullable=True, default=None)



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['stratosphere',
 'stratosphere.store',
 'stratosphere.store.db',
 'stratosphere.utils']

package_data = \
{'': ['*']}

install_requires = \
['cloudpickle>=2.2.0,<3.0.0',
 'colorama>=0.4.6,<0.5.0',
 'joblib>=1.2.0,<2.0.0',
 'pandas>=1.5.1,<2.0.0',
 'sqlalchemy-utils>=0.38.3,<0.39.0',
 'sqlalchemy>=1.4.44,<2.0.0',
 'tqdm>=4.64.1,<5.0.0',
 'ulid-py>=1.1.0,<2.0.0']

extras_require = \
{'complete': ['tabulate>=0.9.0,<0.10.0',
              'ipywidgets>=8.0.2,<9.0.0',
              'scikit-learn>=1.1.3,<2.0.0',
              'dask[complete]>=2022.11.0,<2023.0.0'],
 'dask': ['dask[complete]>=2022.11.0,<2023.0.0'],
 'pgsql': ['psycopg2-binary>=2.9.5,<3.0.0']}

setup_kwargs = {
    'name': 'stratosphere',
    'version': '0.1.54',
    'description': 'A lightweight experimentation toolkit for data scientists.',
    'long_description': '# Stratosphere\n\n*A lightweight experimentation toolkit for data scientists.*\n\n![PyPI - Python Version](https://img.shields.io/pypi/pyversions/stratosphere)\n![PyPI - License](https://img.shields.io/pypi/l/stratosphere)\n![PyPI - Version](https://img.shields.io/pypi/v/stratosphere)\n![PyPI - Wheel](https://img.shields.io/pypi/wheel/stratosphere)\n![PyPI - Installs](https://img.shields.io/pypi/dm/stratosphere)\n![Black - Code style](https://img.shields.io/badge/code%20style-black-000000.svg)\n[![Open in Colab](https://colab.research.google.com/assets/colab-badge.svg)](https://colab.research.google.com/drive/1dkKBwhm4L_MMoWWtfD0FAFgTFP1BV40c)\n[![Open in Binder](https://mybinder.org/badge_logo.svg)](https://mybinder.org/v2/gh/elehcimd/stratosphere/HEAD)\n\nDesigned to be accessibile, fast and robust. `stratosphere` lets you:\n\n1. **Define** programmatically your experiments\n2. **Execute** them in parallel with different backends\n3. **Track** their real-time metrics and final results\n4. **Store** them as serialized objects and tabular data in your database(s)\n5. **Query** them with the best-suited interface: SQL, Pandas and Python\n\nHow is it possible?\n\n* **Fast and light**: [Try it now with JupyterLite](https://github.com/elehcimd/stratosphere/blob/main/doc/JUPYTERLITE.md), it runs entirely in the browser\n* **Built on top of solid components**: [SQLAlchemy](https://www.sqlalchemy.org/), [SQLite](https://www.sqlite.org/), [Pandas](https://pandas.pydata.org/), [Joblib](https://joblib.readthedocs.io/en/latest/) and [Dask](https://www.dask.org/)\n* **Designed for data accessibility**: Use it where you need it. For example, you can experiment in-memory, persist locally, replicate tables on [render.com](https://render.com), and publish results on [preset.io](https://preset.io) (all free tiers!)\n\n![Stratosphere](https://raw.githubusercontent.com/elehcimd/stratosphere/b6993093ae617b98bcabf5d1d3153a7c3e1383a5/logo.png)\n\n## Installation\n\nOn `Python >= 3.8.0`:\n\n* With PyPI: `pip install "stratosphere[complete]" --upgrade` # Install everything\n* With Poetry: `poetry add stratosphere@latest --extras complete` # Install everything\n\nOn `Python 3.7.13-15` with [Google Colab](https://colab.research.google.com/) or \n[Binder](https://mybinder.org), execute the following at the beginning of the notebook:\n\n```\n# Install dependencies\n!pip install pandas joblib sqlalchemy sqlalchemy-utils ulid-py cloudpickle colorama tqdm --upgrade\n\n# Install [complete] extras, required to run the tutorial\n!pip install tabulate scikit-learn dask[complete] --upgrade\n\n# Install the latest stratosphere version, ignoring the python version and dependencies\n!pip install stratosphere --upgrade --ignore-requires-python --no-dependencies\n\n# Resolve ContextualVersionConflict errors, due to broken dependencies\nimport importlib\nimport pkg_resources\nimportlib.reload(pkg_resources)\n```\n\nYou can also install only the `stratosphere` library (dropping the `[complete]` extras).\nModules like `stratosphere.utils.dask`, `stratosphere.utils.metrics`, and `stratosphere.utils.widgets`\nwon\'t work until you also install `dask[complete]`, `scikit-learn`, and `ipywidgets`, respectively.\nThese extras are required to run some of the tutorial notebooks.\n\n## Documentation\n\n* In the [intro notebook](https://github.com/elehcimd/stratosphere/blob/main/doc/JUPYTERLITE.md), you\'ll be exposed to the key concepts. It runs entirely in the browser.\n* In the [tutorial notebooks](./notebooks/), you\'ll cover the advanced topics. Working locally, on Colab and on Binder.\n\n\n## Project pages\n\n* [PyPI](https://pypi.org/project/stratosphere/)\n* [Github](https://github.com/elehcimd/stratosphere)\n\n## License\n\nThis project is licensed under the terms of the [BSD 3-Clause License](https://github.com/elehcimd/stratosphere/blob/main/LICENSE).\n\n## Development\n\nSee the [development](https://github.com/elehcimd/stratosphere/blob/main/docs/DEVELOPMENT.md) page.\n\n## Contributing\n\nWork in progress!\n',
    'author': 'Michele Dallachiesa',
    'author_email': 'michele.dallachiesa@sigforge.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/elehcimd/stratosphere',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.8.0,<3.11',
}


setup(**setup_kwargs)



import platform
import sys
from types import ModuleType
from typing import Callable, Dict

import pkg_resources

catch_exceptions = (FileNotFoundError, AttributeError)


def is_colab() -> bool:
    """Detect if in Google Colab

    Returns:
        bool: True if in Colab
    """
    return "google.colab" in sys.modules


def is_pyodide() -> bool:
    """Detect if in Pyodide (JupyterLite, ...)

    Returns:
        bool: True if in Pyodide
    """
    return "pyodide" in sys.modules


def get_version_stratosphere() -> str:
    """Return the version of Stratosphere. E.g., "0.1.20".

    Returns:
        str: Version of this package
    """
    return pkg_resources.get_distribution("stratosphere").version


def get_version_packages() -> Dict:
    """Return the version of the installed packages. E.g., {"colorama": "1.2.3", ....}

    Returns:
        _type_: Dict of installed package versions
    """
    return {pkg.key: pkg.version for pkg in pkg_resources.working_set}


def try_callable(func: Callable) -> str:
    """Try to call the function. If it fails, return an empty string.

    Args:
        func (Callable): Function to call.

    Returns:
        str: Output of the function, if available.
    """
    try:
        value = func()
        if value is None:
            return ""
        else:
            return value
    except catch_exceptions:
        return ""


def try_module(module: ModuleType, method_name: str) -> str:
    """Try to call a module function. If  it fails, return an empty string.

    Args:
        module (ModuleType): Module
        method_name (str): Method name

    Returns:
        str: Output of the function, if available.
    """
    try:
        value = getattr(module, method_name)
        if callable(value):
            return try_callable(value)
        else:
            return value
    except catch_exceptions:
        return ""


# Not including uname, as it might disclose part of the network name address, and this could be a security issue.
def get_environment() -> Dict:
    """Return a summary of the Python environment, including installed packages, Python, platform, system versions.

    Returns:
        Dict: An overview of the Python environment.
    """
    return {
        "pkg_version": try_callable(get_version_stratosphere),
        "pkgs_version": try_callable(get_version_packages),
        "platform": {
            "machine": try_module(platform, "machine"),
            "mac_ver": try_module(platform, "mac_ver"),
            "libc_ver": try_module(platform, "libc_ver"),
            "system": try_module(platform, "system"),
            "release": try_module(platform, "release"),
            "java_ver": try_module(platform, "java_ver"),
            "win32_ver": try_module(platform, "win32_ver"),
            "win32_edition": try_module(platform, "win32_edition"),
            "linux_distribution": try_module(platform, "linux_distribution"),  # available in Python 3.7
            "freedesktop_os_release": try_module(platform, "freedesktop_os_release"),  # available in Python 3.10
            "uname": try_module(platform, "uname"),
            "sys_version": try_module(sys, "version"),
            "is_colab": is_colab(),
            "is_pyodide": is_pyodide(),
        },
    }
